""" Configure a logger using a dictionary
"""

import logging.config

# Formatters: change what gets logged
minimal = '%(message)s'
detailed  = '%(asctime)s | %(process)d | %(levelname)s | %(name)s | %(message)s '
formatters = { 
    'minimal': {'format': minimal},
    'detailed': {'format': detailed}
}

class Logger(object):
    """ Get logger and set console and file outputs

        Ex: 
        ```
        from logger import Summary
        log = Logger('summary.log').get_logger('mylogger')
         
        ```
    """
    def __init__(self, 
        file_name='', file_format='detailed', file_mode='w',
        console=True, console_format='detailed', console_level='DEBUG'):
    
        # Handlers: change file and console logging details
        handlers = {}
        if console:
            assert console_format in formatters.keys(), \
                f'Must select formatting type from {list(formatters.keys())}'

            handlers['console_handle'] = { 
                'class': 'logging.StreamHandler',
                'level': 'DEBUG',
                'formatter': console_format,
            }

        if file_name:
            assert type(file_name) is str, 'Must provide name for file logging'
            assert file_format in formatters.keys(), \
                f'Must select formatting type from {list(formatters.keys())}'

            handlers['file_handle'] = { 
                'class': 'logging.FileHandler',
                'level': 'INFO',
                'formatter': file_format,
                'filename': file_name,
                'mode': file_mode
            }
        
        # Loggers: change logging options for root and other packages
        loggers = {
            # Root logger
            '': { 
                'handlers': list(handlers.keys()),
                'level': 'DEBUG',
                'propagate': True
            },
            # External loggers
            'requests': {'level': 'WARNING'},
            'urllib3': {'level': 'WARNING'},
            'matplotlib': {'level': 'WARNING'},
            'chardet.charsetprober': {'level': 'INFO'},
            'parso': {'level': 'INFO'} # Fix for ipython autocomplete bug
        }

        self.log_config = { 
            'version': 1,
            'disable_existing_loggers': False,
            'formatters': formatters,
            'handlers': handlers,
            'loggers': loggers
        }
        
    def start(self, name=__name__):
        logging.config.dictConfig(self.log_config)
        return logging.getLogger(name)



""" General network utility functions
"""

import pandas as pd
import networkx as nx

def set_node_attributes(g, root):
    """Add node attributes (inplace operation)"""
    set_attr = nx.set_node_attributes
    set_attr(g, dict(g.degree()), 'k')
    set_attr(g, dict(g.in_degree()), 'k_in')
    set_attr(g, dict(g.out_degree()), 'k_out')
    set_attr(g, nx.degree_centrality(g), 'degree_centrality')
    set_attr(g, nx.betweenness_centrality(g), 'betweenness_centrality')
    set_attr(g, nx.closeness_centrality(g), 'closeness_centrality')
    set_attr(g, nx.clustering(nx.Graph(g)), 'clustering')
    set_attr(g, nx.single_source_shortest_path_length(g, root), 'network_depth')

def set_edge_attributes(g):
    set_attr = nx.set_edge_attributes
    set_attr(g, 'betweenness_centrality', nx.edge_betweenness_centrality(g))

def nodes_to_df(g):
    """Convert nodes dictionary to df index=nodes, cols=attr"""
    node_df = pd.Series(dict(g.nodes())).apply(pd.Series)
    return node_df.reset_index().rename(columns={'index':'node'})

def get_root_component(g, root):
    # Get weakly connected component containing root node
    weak_subgraphs = (g.subgraph(c) for c in nx.weakly_connected_components(g))
    for subgraph in weak_subgraphs:
        if subgraph.has_node(root):
            return subgraph

def find_unreachable_nodes(g, root):
    # Find isolated nodes
    has_depth = nx.single_source_shortest_path_length(g, root).keys()
    return [n for n in g.nodes if n not in has_depth]



""" Recursively retrieve autocomplete suggestions from Google and Bing.
"""

import time
import json
import urllib
import requests
from numpy import random
from datetime import datetime

from . import parsing
from . import logger
log = logger.Logger().start()

##-----------------------------------------------------------------------------
##  Scraper
##-----------------------------------------------------------------------------

def sleep_random(x=0.7, y=1.4):
    """Sleep a random time with noise between x and y seconds"""
    time.sleep(random.uniform(x,y))
    
def prepare_qry(qry):
    return urllib.parse.quote_plus(qry)

def get_google_url():
    return 'https://www.google.com/complete/search?sclient=psy-ab&hl=it&q='

def get_bing_url(cvid='&cvid=CF23583902D944F1874B7D9E36F452CD'):
    return f'http://www.bing.com/AS/Suggestions?&mkt=it-it{cvid}&q='

def scraper(qry, source='bing', sesh=None, sleep=None, allow_zip=False):
    """Scraper with logging and specified user agent
    
    Parameters
    ----------
    qry : str
        Search query to submit
    source : str
        Search engine to submit query to, either "bing" or "google"
    sesh : str, optional
        Pass a custom requests session
    allow_zip : bool, optional
        Boolean controlling whether to unzip response content before parsing.
        Will not work without enabling zip requests in the headers and passing
        the resulting session. To enable zip requests, you must set a header 
        of 'Accept-Encoding' to 'gzip, deflate, br'. This will lighten the load 
        on the servers you are submitting to.

    Returns
    -------
    request.reponse
        Either an HTML (bing) or a JSON (google) response based on source
    """
    assert source in ['bing','google'], "Must select bing or google as source"

    sesh = sesh if sesh else requests.Session()

    base = get_bing_url() if source == 'bing' else get_google_url() 
    url = base + prepare_qry(qry)

    # Sleep
    time.sleep(sleep) if sleep else sleep_random()
    log.info('%s | %s', '%s' % source, qry)
    try:
        response = sesh.get(url, timeout=10)
        if source == 'google':
            return json.loads(response.content)
        elif source == 'bing':
            return response.content.decode('utf-8')
    except Exception as e:
        log.exception('ERROR SCRAPING: request[%s]', response.status_code)
        pass


##-----------------------------------------------------------------------------
##  Get Suggestions
##-----------------------------------------------------------------------------

def get_suggests(qry, source='bing', sesh=None, sleep=None):
    """ Scrape and parse search engine suggestion data for a query.
    
    Parameters
    ----------
        qry : str
            Query to obtain suggestions for.
            
        source : str
            The search engine to submit the query to.

        sesh : requests.Session()
            Pass session to specify headers or maintain connection while building trees.
    """
    sesh = sesh if sesh else requests.Session()
    
    tree = {}
    tree['qry'] = qry
    tree['datetime'] = str(datetime.utcnow())
    tree['source'] = source
    tree['data'] = scraper(qry, source, sesh, sleep)
    
    # Attempt parsing
    parser = parsing.parse_bing if source == 'bing' else parsing.parse_google
    parsed = parser(tree['data'], qry)
    tree.update(parsed)
    return tree


# Suggestions tree
def get_suggests_tree(root, source='bing', max_depth=3, save_to='', sesh=None,
    crawl_id=None, sleep=None):
    """Retrieve autocomplete suggestions tree for a root query
    
    Parameters
    ----------
        root : str
            Query to obtain a suggestion tree for.
            
        source : str
            The search engine to submit the query to
            
        max_depth : int
            Maximum breadth first steps from root

        save_to : str
            Optional, filepath append results as json lines

        sesh : requests.Session()
            Pass session to specify headers and maintain connection
    """
    sesh = sesh if sesh else requests.Session()

    depth = 0
    root_branch = get_suggests(root, source, sesh, sleep)
    root_branch['depth'] = depth
    root_branch['root'] = root
    root_branch['crawl_id'] = crawl_id

    if save_to:
        outfile = open(save_to, 'a+')
        outdata = json.dumps(root_branch)
        outfile.write(f'{outdata}\n')

    # Initialize list of suggestion dicts for output
    tree = [root_branch]

    # Initialize tracking of all queries seen
    all_suggests = {root}

    # Recurse
    while depth < max_depth:
        suggests = {d['qry']: d['suggests'] for d in tree if d['depth']==depth}
        depth += 1
        
        for qry, suggest_list in suggests.items():
            if suggest_list:
                for s in suggest_list:
                    if s not in all_suggests: # Don't crawl self-loops or duplicates
                        branches = get_suggests(s, source, sesh, sleep)                    
                        branches['depth'] = depth
                        branches['root'] = root
                        branches['crawl_id'] = crawl_id
                        if save_to: 
                            outfile.write(f'{json.dumps(branches)}\n')
                        tree.append(branches)            
                        all_suggests.add(s)
                        
    if save_to: outfile.close()
    return tree



import re
import os
import json
import random, itertools
import pandas as pd
pd.set_option('max.column', 99)

from timeit import default_timer
from string import ascii_letters, digits

# Files ------------------------------------------------------------------------

def all_abs_paths(dir):
    file_paths = []
    for folder, subs, files in os.walk(dir):
        for filename in files:
            file_paths.append(os.path.abspath(os.path.join(folder, filename)))
    return file_paths

def read_lines(fp):
    with open(fp, 'r') as infile:
        if '.txt' in fp: return [line.strip() for line in infile]
        elif '.json' in fp:  return [json.loads(line.strip()) for line in infile]

def write_lines(iter_data, fp, overwrite=False):
    mode = 'w' if overwrite else 'a+'
    with open(fp, mode) as outfile:
        if '.txt' in fp:
            for data in iter_data:
                outfile.write('%s\n' % data)
        elif '.json' in fp:
            for data in iter_data:
                outfile.write('%s\n' % json.dumps(data))

# Descriptive Stats ------------------------------------------------------------

def dnum(df_col):
    if df_col.dtype in [int, float]:
        return df_col.describe()
    else:
        return 0

def dft(dfcol):
    tab = pd.DataFrame({
        'n': dfcol.value_counts(),
        'p': dfcol.value_counts(True)
    }).sort_index()
    return tab

# Lists ------------------------------------------------------------------------

def unlist(nested_list):
    return list(itertools.chain.from_iterable(nested_list))

# Strings ----------------------------------------------------------------------

def split_by_spaces(s, n=2):
    # Split a string by n or more spaces
    return re.split('\s{%d,}' % n, s)

def get_between_brackets(s, regex=r'\[(.*?)\]'):
    return re.search(regex, s).group(1)

def get_between_parentheses(s, regex=r'\((.*?)\)'):
    return re.search(regex, s).group(1)
    
def remove_digits(string):
    return "".join([x for x in string if not x.isdigit()]).strip()

# Timer ------------------------------------------------------------------------

def start_timer():
    return default_timer()

def stop_timer(start, v=False):
    stop = default_timer()
    total_time = stop - start
    mins, secs = divmod(total_time, 60)
    hours, mins = divmod(mins, 60)
    runtime = f'{int(hours)}:{int(mins)}:{secs}'
    if v:
        print(f'Run time:\t{runtime}\n')
    return f'{runtime}'

# Misc -------------------------------------------------------------------------

def alphanumerics():
    """Generate upper and lowercase letters and digits"""
    return ascii_letters + digits

def random_string(length=12):
    """Generate a random string of alphanumerics"""
    return ''.join(random.choice(alphanumerics()) for i in range(length))

def write_latex_df(df, filename, index=False):
    """Write a latex table from a pd.DataFrame"""
    latex = df.to_latex(index=index)
    with open(filename, 'w') as f:
        f.write(latex)
    print('Saved as: ' + filename)


__version__ = "0.1.1"

from .suggests import get_suggests
from .suggests import get_suggests_tree

from .parsing import parse_bing
from .parsing import parse_google
from .parsing import to_edgelist

from .parsing import add_parent_nodes
from .parsing import add_metanodes



import setuptools
setuptools.setup(
    name="suggests",
    version="0.1.12",
    author='Ronald E. Robertson',
    author_email='rer@ronalderobertson.com',
    description="Algorithm auditing tools for search engine autocomplete",
    keywords="suggestions autocomplete google bing",
    url='http://github.com/gitronald/suggests',
    classifiers=['Programming Language :: Python :: 3',
                 'License :: OSI Approved :: MIT License'],
    packages=setuptools.find_packages(),
    install_requires=['requests', 'pandas', 'numpy', 'bs4'],
    python_requires='~=3.6',
    package_data={'': ['*.txt', '*.md']},
    license='MIT'
)



""" Parsing functions for Google and Bing
"""

import re
import html
import pandas as pd
import urllib.parse
from bs4 import BeautifulSoup
from collections import OrderedDict

from . import logger
log = logger.Logger().start(__name__)

def get_source_target_columns(edges):
    edge_details = edges.edge.apply(pd.Series)
    edge_details.columns = ['source','target']
    return pd.concat([edges, edge_details], axis=1)

def parse_raw_data(raw_data, source):
    # Parse raw aggregated data from `clean` into edge lists
    parser = parse_google if source=='google' else parse_bing
    return pd.concat([raw_data, raw_data.data.apply(parser).apply(pd.Series)], axis=1)
    
def get_edges(data):
    clean_data = [parse_raw_data(df, k) for k, df in data.groupby('source')]
    return pd.concat([to_edgelist(df) for df in clean_data])

def strip_html(string):
    """Strips HTML <tags>"""
    return re.sub('<[^<]+?>', '', string)

def parse_google(json_data, qry=''):
    
    def google_parser(json_data):

        def suggest_parser(s):
            return html.unescape(strip_html(s))

        qry = json_data[0]
        suggests = [s[0]+' - '+s[3]['b'] if s[1]==46 else s[0] for s in json_data[1]]
        suggests = [suggest_parser(s) for s in suggests]
        tags = json_data[2]
        return qry, suggests, tags
    
    try:
        qry, suggests, tags = google_parser(json_data)
        self_loops = [i for i, s in enumerate(suggests) if s == qry]
        parsed = {'suggests': suggests, 'self_loops':self_loops, 'tags': tags}
    except:
        log.exception('ERROR PARSING GOOGLE:\n%s', json_data)
        parsed = {'suggests':[], 'self_loops':[], 'tags':[]}
    return parsed

def parse_bing_qry(raw_html, qry=''):
    """Recover query from bing response html"""
    url = BeautifulSoup(raw_html).find('li')['url']
    if url:
        return str(urllib.parse.parse_qs(urllib.parse.urlparse().query)['pq'][0])
    else:
        return None

def parse_bing(raw_html, qry=''):
    
    def bing_parser(raw_html):
        soup = BeautifulSoup(raw_html, 'html.parser')
        if not soup.text:
            # No suggestions
            return []
        suggests = [div.text for div in soup.find_all('div', {'class':'sa_tm'})]
        suggests = [html.unescape(s) for s in suggests]
        return suggests

    try:
        suggests = bing_parser(raw_html)
        self_loops = [i for i, s in enumerate(suggests) if s == qry]
        parsed = {'suggests':suggests, 'self_loops':self_loops, 'tags':[]}
    except:
        log.exception('ERROR PARSING BING:\n%s', raw_html)
        parsed = {'suggests':[], 'self_loops':[], 'tags':[]}
    return parsed

def to_edgelist(tree, self_loops=False):
    """Convert suggestions tree (df) to edgelist (df)"""
    edge_list = []
    assert type(tree) == list or type(tree) == pd.np.recarray, \
        "Must pass records or a list of dicts"

    for row in tree:

        if self_loops:
            suggests = row['suggests'] 
        else:
            suggests = [s for s in row['suggests'] if s != row['qry']]
            
        if suggests:
            for rank, s in enumerate(suggests):
                edge = OrderedDict([
                    ('root', row['root']), 
                    ('edge', (row['qry'], s)), 
                    ('source', row['qry']),
                    ('target', html.unescape(s)),
                    ('rank', rank + 1), 
                    ('depth', row['depth']),
                    ('search_engine', row['source']),
                    ('datetime', row['datetime'])
                ])
                edge_list.append(edge)
        else: # If no suggests at root, append empty root
            if row['depth'] == 0:
                no_edges = OrderedDict([
                    ('root', row['root']), 
                    ('edge', None), 
                    ('source', row['qry']),
                    ('target', None),
                    ('rank', 1), 
                    ('depth', row['depth']),
                    ('search_engine', row['source']),
                    ('datetime', row['datetime'])
                ])
                edge_list.append(no_edges)
                
    edge_df = pd.DataFrame(edge_list)
    return edge_df

def add_parent_nodes(edges):
    edges_original = edges.copy()

    # Get parent node
    parent = edges.rename(columns={"source": "parent", "target": "source"})
    parent["depth"] = parent["depth"] + 1
    parent = parent[["root", "parent", "source", "depth", "search_engine"]]
    edges = edges.merge(parent, on = ["root", "source", "depth", "search_engine"], how = "left")

    # Get grandparent node
    grandparent = edges.rename(columns={"parent": "grandparent", "source": "parent", "target": "source"})
    grandparent["depth"] = grandparent["depth"] + 1
    grandparent = grandparent[["root", "grandparent", "parent", "source", "depth", "search_engine"]]
    edges = edges.merge(grandparent, on = ["root", "parent", "source", "depth", "search_engine"], how = "left")

    # Resolve merge points
    gb = edges.groupby('edge')
    merged_parents = pd.DataFrame({
        'grandparent': gb.grandparent.apply(lambda col: col.str.cat(sep=' ') if col.any() else None),
        'parent': gb.parent.apply(lambda col: col.str.cat(sep=' ') if col.any() else None)
    }).reset_index()
    edges = edges_original.merge(merged_parents, on='edge', how='left')
    return edges

def add_metanodes(row):
    # Get memory from two steps back
    grandparent = list() if row.isna()["grandparent"] else row["grandparent"].split(" ")
    parent = list() if row.isna()["parent"] else row["parent"].split(" ")

    # Set current row source and target nodes
    source = row["source"].split(" ")
    target = row["target"].split(" ")

    # Calculate differences 
    source_add = [i for i in source if i not in set(parent)]
    target_add = [i for i in target if i not in set(source)]

    # Track difference in previous step
    parent_add = [i for i in parent if i not in set(grandparent)]

    if not source_add: # information removed
        source_add = parent_add
    # If target adds nothing nothing, circle back
    if not target_add: 
        print(f'circle back: {source_add}')
        target_add = source_add

    row["source_add"] = " ".join(source_add)
    row["target_add"] = " ".join(target_add)
    return row
